#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Resolve targets for `CandidateDevice`.

Resolving a target involves looking up an `MbedTarget` from the `mbed-targets` API, using data found in the "htm file"
located on an "Mbed Enabled" device's USB MSD.

For more information on the mbed-targets package visit https://github.com/ARMmbed/mbed-targets
"""
import logging

from typing import Optional

from mbed_tools.targets import (
    Board,
    get_board_by_product_code,
    get_board_by_online_id,
    get_board_by_jlink_slug,
)
from mbed_tools.targets.exceptions import UnknownBoard, MbedTargetsError

from mbed_tools.devices._internal.exceptions import NoBoardForCandidate, ResolveBoardError
from mbed_tools.devices._internal.file_parser import OnlineId


logger = logging.getLogger(__name__)


def resolve_board(
    product_code: Optional[str] = None, online_id: Optional[OnlineId] = None, serial_number: str = ""
) -> Board:
    """Resolves a board object from the platform database.

    We have multiple ways to identify boards from various metadata sources Mbed provides. This is because there are
    many supported Mbed device families, each with slightly different ways of identifying themselves as Mbed enabled.
    Because of this we need to try each input in turn, falling back to the next lookup method in the priority order if
    the previous one was unsuccessful.

    The rules are as follows:

    1. Use the product code from the mbed.htm file or details.txt if available
    2. Use online ID from the htm file or Board.html if available
    3. Try to use the first 4 chars of the USB serial number as the product code
    """
    if product_code:
        try:
            return get_board_by_product_code(product_code)
        except UnknownBoard:
            logger.error(f"Could not identify a board with the product code: '{product_code}'.")
        except MbedTargetsError as e:
            logger.error(
                f"There was an error looking up the product code `{product_code}` from the target database.\nError: {e}"
            )
            raise ResolveBoardError() from e

    if online_id:
        slug = online_id.slug
        target_type = online_id.target_type
        try:
            if target_type == "jlink":
                return get_board_by_jlink_slug(slug=slug)
            else:
                return get_board_by_online_id(slug=slug, target_type=target_type)
        except UnknownBoard:
            logger.error(f"Could not identify a board with the slug: '{slug}' and target type: '{target_type}'.")
        except MbedTargetsError as e:
            logger.error(
                f"There was an error looking up the online ID `{online_id!r}` from the target database.\nError: {e}"
            )
            raise ResolveBoardError() from e

    # Product code might be the first 4 characters of the serial number
    product_code = serial_number[:4]
    if product_code:
        try:
            return get_board_by_product_code(product_code)
        except UnknownBoard:
            # Most devices have a serial number so this may not be a problem
            logger.info(
                f"The device with the Serial Number: '{serial_number}' (Product Code: '{product_code}') "
                f"does not appear to be an Mbed development board."
            )
        except MbedTargetsError as e:
            logger.error(
                f"There was an error looking up the product code `{product_code}` from the target database.\nError: {e}"
            )
            raise ResolveBoardError() from e

    raise NoBoardForCandidate



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Command to generate the application CMake configuration script used by the build/compile system."""
import pathlib

import click

from mbed_tools.project import MbedProgram
from mbed_tools.build import generate_config


@click.command(
    help="Generate an Mbed OS config CMake file and write it to a .mbedbuild folder in the program directory."
)
@click.option(
    "--custom-targets-json", type=click.Path(), default=None, help="Path to custom_targets.json.",
)
@click.option(
    "-t",
    "--toolchain",
    type=click.Choice(["ARM", "GCC_ARM"], case_sensitive=False),
    required=True,
    help="The toolchain you are using to build your app.",
)
@click.option("-m", "--mbed-target", required=True, help="A build target for an Mbed-enabled device, eg. K64F")
@click.option("-b", "--profile", default="develop", help="The build type (release, develop or debug).")
@click.option("-o", "--output-dir", type=click.Path(), default=None, help="Path to output directory.")
@click.option(
    "-p",
    "--program-path",
    type=click.Path(),
    default=".",
    help="Path to local Mbed program. By default is the current working directory.",
)
@click.option(
    "--mbed-os-path", type=click.Path(), default=None, help="Path to local Mbed OS directory.",
)
@click.option(
    "--app-config", type=click.Path(), default=None, help="Path to application configuration file.",
)
def configure(
    toolchain: str,
    mbed_target: str,
    profile: str,
    program_path: str,
    mbed_os_path: str,
    output_dir: str,
    custom_targets_json: str,
    app_config: str
) -> None:
    """Exports a mbed_config.cmake file to build directory in the program root.

    The parameters set in the CMake file will be dependent on the combination of
    toolchain and Mbed target provided and these can then control which parts of
    Mbed OS are included in the build.

    This command will create the .mbedbuild directory at the program root if it doesn't
    exist.

    Args:
        custom_targets_json: the path to custom_targets.json
        toolchain: the toolchain you are using (eg. GCC_ARM, ARM)
        mbed_target: the target you are building for (eg. K64F)
        profile: The Mbed build profile (debug, develop or release).
        program_path: the path to the local Mbed program
        mbed_os_path: the path to the local Mbed OS directory
        output_dir: the path to the output directory
        app_config: the path to the application configuration file
    """
    cmake_build_subdir = pathlib.Path(mbed_target.upper(), profile.lower(), toolchain.upper())
    if mbed_os_path is None:
        program = MbedProgram.from_existing(pathlib.Path(program_path), cmake_build_subdir)
    else:
        program = MbedProgram.from_existing(pathlib.Path(program_path), cmake_build_subdir, pathlib.Path(mbed_os_path))
    if custom_targets_json is not None:
        program.files.custom_targets_json = pathlib.Path(custom_targets_json)
    if output_dir is not None:
        program.files.cmake_build_dir = pathlib.Path(output_dir)
    if app_config is not None:
        program.files.app_config_file = pathlib.Path(app_config)

    mbed_target = mbed_target.upper()
    _, output_path = generate_config(mbed_target, toolchain, program)
    click.echo(f"mbed_config.cmake has been generated and written to '{str(output_path.resolve())}'")



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import json
import pathlib
import pytest
from click.testing import CliRunner
from mbed_tools.devices.device import ConnectedDevices
from mbed_tools.targets import Board
from unittest import mock

from mbed_tools.cli.list_connected_devices import list_connected_devices
from mbed_tools.devices import Device


@pytest.fixture
def get_connected_devices():
    with mock.patch("mbed_tools.cli.list_connected_devices.get_connected_devices") as cd:
        yield cd


def create_fake_device(
    serial_number="8675309",
    serial_port="/dev/ttyUSB/default",
    mount_points=(pathlib.Path("/media/you/DISCO"), pathlib.Path("/media/you/NUCLEO")),
    interface_version="VA3JME",
    board_type="BoardType",
    board_name="BoardName",
    product_code="0786",
    target_type="TargetType",
    slug="slug",
    build_variant=("NS", "S"),
    mbed_os_support=("mbed1", "mbed2"),
    mbed_enabled=("baseline", "extended"),
):
    device_attrs = {
        "serial_number": serial_number,
        "serial_port": serial_port,
        "mount_points": mount_points,
        "interface_version": interface_version,
    }
    board_attrs = {
        "board_type": board_type,
        "board_name": board_name,
        "product_code": product_code,
        "target_type": target_type,
        "slug": slug,
        "build_variant": build_variant,
        "mbed_os_support": mbed_os_support,
        "mbed_enabled": mbed_enabled,
    }
    return Device(Board(**board_attrs), **device_attrs)


class TestListConnectedDevices:
    def test_informs_when_no_devices_are_connected(self, get_connected_devices):
        get_connected_devices.return_value = ConnectedDevices()

        result = CliRunner().invoke(list_connected_devices)

        assert result.exit_code == 0
        assert "No connected Mbed devices found." in result.output


class TestListConnectedDevicesTabularOutput:
    @pytest.mark.parametrize(
        "header, device_attr",
        [
            ("Serial number", "serial_number"),
            ("Serial port", "serial_port"),
            ("Interface Version", "interface_version"),
        ],
    )
    def test_device_attr_included(self, header, device_attr, get_connected_devices):
        heading_name = header
        device = create_fake_device()
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])
        attr = getattr(device, device_attr)

        result = CliRunner().invoke(list_connected_devices)

        output_lines = result.output.splitlines()

        assert result.exit_code == 0
        assert heading_name in output_lines[0]
        assert attr in output_lines[2]
        assert output_lines[0].find(heading_name) == output_lines[2].find(attr)

    @pytest.mark.parametrize("header, board_attr", [("Board name", "board_name")])
    def test_board_attr_included(self, header, board_attr, get_connected_devices):
        heading_name = header
        device = create_fake_device()
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])
        attr = getattr(device.mbed_board, board_attr)

        result = CliRunner().invoke(list_connected_devices)

        output_lines = result.output.splitlines()
        assert result.exit_code == 0
        assert heading_name in output_lines[0]
        assert attr in output_lines[2]
        assert output_lines[0].find(heading_name) == output_lines[2].find(attr)

    @pytest.mark.parametrize("header, device_attr", [("Mount point(s)", "mount_points")])
    def test_multiline_device_attr_included(self, header, device_attr, get_connected_devices):
        heading_name = header
        device = create_fake_device()
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])
        attr = getattr(device, device_attr)

        result = CliRunner().invoke(list_connected_devices)

        output_lines = result.output.splitlines()
        assert result.exit_code == 0
        assert heading_name in output_lines[0]
        for i, a in enumerate(attr):
            assert str(a) in output_lines[2 + i]
            assert output_lines[0].find(heading_name) == output_lines[2 + i].find(str(a))

    def test_build_target_included(self, get_connected_devices):
        heading_name = "Build target(s)"
        device = create_fake_device()
        board = device.mbed_board
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])
        build_targets = [
            f"{board.board_type}_{board.build_variant[0]}",
            f"{board.board_type}_{board.build_variant[1]}",
            f"{board.board_type}",
        ]

        result = CliRunner().invoke(list_connected_devices)

        output_lines = result.output.splitlines()
        assert result.exit_code == 0
        assert heading_name in output_lines[0]
        for i, bt in enumerate(build_targets):
            assert bt in output_lines[2 + i]
            assert output_lines[0].find(heading_name) == output_lines[2 + i].find(bt)

    def test_appends_identifier_when_identical_boards_found(self, get_connected_devices):
        heading_name = "Build target(s)"
        device_a = create_fake_device()
        device_b = create_fake_device()
        board = device_a.mbed_board
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device_a, device_b])
        build_targets = [
            f"{board.board_type}_{board.build_variant[0]}[0]",
            f"{board.board_type}_{board.build_variant[1]}[0]",
            f"{board.board_type}[0]",
            f"{board.board_type}_{board.build_variant[0]}[1]",
            f"{board.board_type}_{board.build_variant[1]}[1]",
            f"{board.board_type}[1]",
        ]

        result = CliRunner().invoke(list_connected_devices)

        output_lines = result.output.splitlines()
        assert result.exit_code == 0
        assert heading_name in output_lines[0]
        for i, bt in enumerate(build_targets):
            assert bt in output_lines[2 + i]
            assert output_lines[0].find(heading_name) == output_lines[2 + i].find(bt)

    def test_sort_order(self, get_connected_devices):
        devices = [
            create_fake_device(
                serial_number="11111111",
                board_name="AAAAAAA1FirstBoard",
                build_variant=(),
                mount_points=(pathlib.Path("/media/you/First"),),
            ),
            create_fake_device(
                serial_number="22222222",
                board_name="SameBoard[0]",
                build_variant=(),
                mount_points=(pathlib.Path("/media/you/Second"),),
            ),
            create_fake_device(
                serial_number="33333333",
                board_name="SameBoard[1]",
                build_variant=(),
                mount_points=(pathlib.Path("/media/you/Third"),),
            ),
        ]
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[devices[2], devices[0], devices[1]])

        result = CliRunner().invoke(list_connected_devices, "--show-all")
        output_lines = result.output.splitlines()
        assert devices[0].mbed_board.board_name in output_lines[2]
        assert devices[1].mbed_board.board_name in output_lines[3]
        assert devices[2].mbed_board.board_name in output_lines[4]

    def test_displays_unknown_serial_port_value(self, get_connected_devices):
        heading_name = "Serial port"
        device = create_fake_device(serial_port=None)
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])

        result = CliRunner().invoke(list_connected_devices)

        output_lines = result.output.splitlines()
        assert result.exit_code == 0
        assert heading_name in output_lines[0]
        assert output_lines[0].find(heading_name) == output_lines[2].find("<unknown>")


class TestListConnectedDevicesJSONOutput:
    @pytest.mark.parametrize("device_attr", ("serial_number", "serial_port", "mount_points", "interface_version"))
    def test_device_attr_included(self, device_attr, get_connected_devices):
        device = create_fake_device()
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])
        attr = getattr(device, device_attr)
        # JSON encoder converts tuples to lists, so we need to convert the test data also to match
        attr = [str(a) for a in attr] if isinstance(attr, tuple) else str(attr)

        result = CliRunner().invoke(list_connected_devices, "--format=json")

        assert result.exit_code == 0
        assert attr == json.loads(result.output)[0][device_attr]

    @pytest.mark.parametrize(
        "board_attr", ("product_code", "board_type", "board_name", "mbed_os_support", "mbed_enabled")
    )
    def test_board_attr_included(self, board_attr, get_connected_devices):
        device = create_fake_device()
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])
        attr = getattr(device.mbed_board, board_attr)
        # JSON encoder converts tuples to lists, so we need to convert the test data also to match
        attr = [str(a) for a in attr] if isinstance(attr, tuple) else str(attr)

        result = CliRunner().invoke(list_connected_devices, "--format=json")

        assert result.exit_code == 0
        assert attr == json.loads(result.output)[0]["mbed_board"][board_attr]

    def test_build_targets_included(self, get_connected_devices):
        device = create_fake_device()
        board = device.mbed_board
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device])

        result = CliRunner().invoke(list_connected_devices, "--format=json")

        expected_output = [
            {
                "build_targets": [
                    f"{board.board_type}_{board.build_variant[0]}",
                    f"{board.board_type}_{board.build_variant[1]}",
                    f"{board.board_type}",
                ],
            },
        ]

        assert result.exit_code == 0
        for actual, expected in zip(json.loads(result.output), expected_output):
            assert actual["mbed_board"]["build_targets"] == expected["build_targets"]

    def test_identifiers_appended_when_identical_boards_found(self, get_connected_devices):
        device = create_fake_device()
        device_2 = create_fake_device()
        board = device.mbed_board
        get_connected_devices.return_value = ConnectedDevices(identified_devices=[device, device_2])

        result = CliRunner().invoke(list_connected_devices, "--format=json")

        expected_output = [
            {
                "build_targets": [
                    f"{board.board_type}_{board.build_variant[0]}[0]",
                    f"{board.board_type}_{board.build_variant[1]}[0]",
                    f"{board.board_type}[0]",
                ],
            },
            {
                "build_targets": [
                    f"{board.board_type}_{board.build_variant[0]}[1]",
                    f"{board.board_type}_{board.build_variant[1]}[1]",
                    f"{board.board_type}[1]",
                ],
            },
        ]

        assert result.exit_code == 0
        for actual, expected in zip(json.loads(result.output), expected_output):
            assert actual["mbed_board"]["build_targets"] == expected["build_targets"]



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Wrappers for git operations."""
from dataclasses import dataclass
from pathlib import Path

import git
import logging

from mbed_tools.project.exceptions import VersionControlError
from mbed_tools.project._internal.progress import ProgressReporter
from typing import Optional

logger = logging.getLogger(__name__)


@dataclass
class GitReference:
    """Git reference for a remote repository.

    Attributes:
        repo_url: URL of the git repository.
        ref: The reference commit sha, tag or branch.
    """

    repo_url: str
    ref: str


def clone(url: str, dst_dir: Path, ref: Optional[str] = None, depth: int = 1) -> git.Repo:
    """Clone a library repository.

    Args:
        url: URL of the remote to clone.
        dst_dir: Destination directory for the cloned repo.
        ref: An optional git commit hash, branch or tag reference to checkout
        depth: Truncate history to the specified number of commits. Defaults to
               1, to make a shallow clone.

    Raises:
        VersionControlError: Cloning the repository failed.
    """
    # Gitpython doesn't propagate the git error message when a repo is already
    # cloned, so we cannot depend on git to handle the "already cloned" error.
    # We must handle this ourselves instead.
    if dst_dir.exists() and list(dst_dir.glob("*")):
        raise VersionControlError(f"{dst_dir} exists and is not an empty directory.")

    clone_from_kwargs = {"url": url, "to_path": str(dst_dir), "progress": ProgressReporter(name=url), "depth": depth}
    if ref:
        clone_from_kwargs["branch"] = ref

    try:
        return git.Repo.clone_from(**clone_from_kwargs)
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Cloning git repository from url '{url}' failed. Error from VCS: {err}")


def checkout(repo: git.Repo, ref: str, force: bool = False) -> None:
    """Check out a specific reference in the given repository.

    Args:
        repo: git.Repo object where the checkout will be performed.
        ref: Git commit hash, branch or tag reference, must be a valid ref defined in the repo.

    Raises:
        VersionControlError: Check out failed.
    """
    try:
        git_args = [ref] + ["--force"] if force else [ref]
        repo.git.checkout(*git_args)
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Failed to check out revision '{ref}'. Error from VCS: {err}")


def fetch(repo: git.Repo, ref: str) -> None:
    """Fetch from the repo's origin.

    Args:
        repo: git.Repo object where the checkout will be performed.
        ref: Git commit hash, branch or tag reference, must be a valid ref defined in the repo.

    Raises:
        VersionControlError: Fetch failed.
    """
    try:
        repo.git.fetch("origin", ref)
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Failed to fetch. Error from VCS: {err}")


def init(path: Path) -> git.Repo:
    """Initialise a git repository at the given path.

    Args:
        path: Path where the repo will be initialised.

    Returns:
        Initialised git.Repo object.

    Raises:
        VersionControlError: initalising the repository failed.
    """
    try:
        return git.Repo.init(str(path))
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Failed to initialise git repository at path '{path}'. Error from VCS: {err}")


def get_repo(path: Path) -> git.Repo:
    """Get a git.Repo object from an existing repository path.

    Args:
        path: Path to the git repository.

    Returns:
        git.Repo object.

    Raises:
        VersionControlError: No valid git repository at this path.
    """
    try:
        return git.Repo(str(path))
    except git.exc.InvalidGitRepositoryError:
        raise VersionControlError(
            "Could not find a valid git repository at this path. Please perform a `git init` command."
        )


def get_default_branch(repo: git.Repo) -> str:
    """Get a default branch from an existing git.Repo.

    Args:
        repo: git.Repo object

    Returns:
        The default branch name as a string.

    Raises:
        VersionControlError: Could not find the default branch name.
    """
    try:
        return str(repo.git.symbolic_ref("refs/remotes/origin/HEAD").rsplit("/", maxsplit=1)[-1])
    except git.exc.GitCommandError as err:
        raise VersionControlError(f"Could not resolve default repository branch name. Error from VCS: {err}")



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from unittest import TestCase, mock

from click.testing import CliRunner

from mbed_tools.cli.configure import configure


class TestConfigureCommand(TestCase):
    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_generate_config_called_with_correct_arguments(self, program, generate_config):
        CliRunner().invoke(configure, ["-m", "k64f", "-t", "gcc_arm"])

        generate_config.assert_called_once_with("K64F", "GCC_ARM", program.from_existing())

    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_generate_config_called_with_mbed_os_path(self, program, generate_config):
        CliRunner().invoke(configure, ["-m", "k64f", "-t", "gcc_arm", "--mbed-os-path", "./extern/mbed-os"])

        generate_config.assert_called_once_with("K64F", "GCC_ARM", program.from_existing())

    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_custom_targets_location_used_when_passed(self, program, generate_config):
        program = program.from_existing()
        custom_targets_json_path = pathlib.Path("custom", "custom_targets.json")
        CliRunner().invoke(
            configure, ["-t", "gcc_arm", "-m", "k64f", "--custom-targets-json", custom_targets_json_path]
        )

        generate_config.assert_called_once_with("K64F", "GCC_ARM", program)
        self.assertEqual(program.files.custom_targets_json, custom_targets_json_path)

    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_custom_output_directory_used_when_passed(self, program, generate_config):
        program = program.from_existing()
        output_dir = pathlib.Path("build")
        CliRunner().invoke(configure, ["-t", "gcc_arm", "-m", "k64f", "-o", output_dir])

        generate_config.assert_called_once_with("K64F", "GCC_ARM", program)
        self.assertEqual(program.files.cmake_build_dir, output_dir)

    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_app_config_used_when_passed(self, program, generate_config):
        program = program.from_existing()
        app_config_path = pathlib.Path("alternative_config.json")
        CliRunner().invoke(
            configure, ["-t", "gcc_arm", "-m", "k64f", "--app-config", app_config_path]
        )

        generate_config.assert_called_once_with("K64F", "GCC_ARM", program)
        self.assertEqual(program.files.app_config_file, app_config_path)

    @mock.patch("mbed_tools.cli.configure.generate_config")
    @mock.patch("mbed_tools.cli.configure.MbedProgram")
    def test_profile_used_when_passed(self, program, generate_config):
        test_program = program.from_existing()
        program.reset_mock()  # clear call count from previous line

        toolchain = "gcc_arm"
        target = "k64f"
        profile = "release"

        CliRunner().invoke(
            configure, ["-t", toolchain, "-m", target, "--profile", profile]
        )

        program.from_existing.assert_called_once_with(
            pathlib.Path("."),
            pathlib.Path(target.upper(), profile, toolchain.upper())
        )
        generate_config.assert_called_once_with("K64F", "GCC_ARM", test_program)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib

from unittest import mock

import pytest

from mbed_tools.devices._internal.file_parser import OnlineId, read_device_files


class TestReadDeviceFiles:
    def test_finds_daplink_compatible_device_files(self, tmp_path):
        details = pathlib.Path(tmp_path, "details.txt")
        htm = pathlib.Path(tmp_path, "mbed.htm")
        htm.write_text("code=2222")
        details.write_text("Version: 2")

        info = read_device_files([tmp_path])

        assert info.product_code is not None
        assert info.interface_details is not None

    def test_finds_jlink_device_files(self, tmp_path):
        segger_html = pathlib.Path(tmp_path, "Segger.html")
        board_html = pathlib.Path(tmp_path, "Board.html")
        segger_html.write_text(build_segger_html())
        board_html.write_text(build_board_html())

        info = read_device_files([tmp_path])

        assert info.online_id is not None
        assert info.interface_details is not None

    def test_warns_if_no_device_files_found(self, caplog, tmp_path):
        read_device_files([tmp_path])

        assert str(tmp_path) in caplog.text

    def test_skips_hidden_files(self, caplog, tmp_path):
        auth = "101000000000000000000002F7F35E602eeb0bb9b632205c51f6c357aeee7bc9"
        file_contents = (
            '<meta http-equiv="refresh" '
            f'content="0; url=http://mbed.org/start?auth={auth}&loader=11972&firmware=16457&configuration=4" />'
        )
        pathlib.Path(tmp_path, "._MBED.HTM").write_text(file_contents)
        pathlib.Path(tmp_path, "._DETAILS.TXT").write_text("Version: 2222")

        assert read_device_files([tmp_path]).product_code is None
        assert not read_device_files([tmp_path]).interface_details

    def test_handles_os_error_with_warning(self, tmp_path, caplog, monkeypatch):
        bad_htm = pathlib.Path(tmp_path, "MBED.HTM")
        bad_htm.touch()
        monkeypatch.setattr(pathlib.Path, "read_text", mock.Mock(side_effect=OSError))

        read_device_files([tmp_path])
        assert str(bad_htm) in caplog.text


class TestExtractProductCodeFromHtm:
    def test_reads_product_code_from_code_attribute(self, tmp_path):
        code = "02400201B80ECE4A45F033F2"
        file_contents = f'<meta http-equiv="refresh" content="0; url=http://mbed.org/device/?code={code}"/>'
        pathlib.Path(tmp_path, "MBED.HTM").write_text(file_contents)

        assert read_device_files([tmp_path]).product_code == code[:4]

    def test_reads_product_code_from_auth_attribute(self, tmp_path):
        auth = "101000000000000000000002F7F35E602eeb0bb9b632205c51f6c357aeee7bc9"
        file_contents = (
            '<meta http-equiv="refresh" '
            f'content="0; url=http://mbed.org/start?auth={auth}&loader=11972&firmware=16457&configuration=4" />'
        )
        pathlib.Path(tmp_path, "MBED.HTM").write_text(file_contents)

        assert read_device_files([tmp_path]).product_code == auth[:4]

    def test_none_if_no_product_code(self, tmp_path):
        file_contents = '<meta http-equiv="refresh" content="0; url=http://mbed.org/config" />'
        pathlib.Path(tmp_path, "MBED.HTM").write_text(file_contents)

        assert read_device_files([tmp_path]).product_code is None

    def test_extracts_first_product_code_found(self, tmp_path):
        auth = "101000000000000000000002F7F35E602eeb0bb9b632205c51f6c357aeee7bc9"
        file_contents_1 = (
            '<meta http-equiv="refresh" '
            f'content="0; url=http://mbed.org/start?auth={auth}&loader=11972&firmware=16457&configuration=4" />'
        )
        code = "02400201B80ECE4A45F033F2"
        file_contents_2 = f'<meta http-equiv="refresh" content="0; url=http://mbed.org/device/?code={code}"/>'
        directory_1 = pathlib.Path(tmp_path, "test-1")
        directory_1.mkdir()
        directory_2 = pathlib.Path(tmp_path, "test-2")
        directory_2.mkdir()
        pathlib.Path(directory_1, "mbed.htm").write_text(file_contents_1)
        pathlib.Path(directory_2, "mbed.htm").write_text(file_contents_2)

        result = read_device_files([directory_1, directory_2])

        assert result.product_code == auth[:4]


class TestExtractOnlineIDFromHTM:
    def test_reads_online_id_from_url(self, tmp_path):
        url = "https://os.mbed.com/platforms/THIS-IS_a_SLUG_123/"
        file_contents = f"window.location.replace({url});"
        pathlib.Path(tmp_path, "MBED.HTM").write_text(file_contents)

        assert read_device_files([tmp_path]).online_id == OnlineId(target_type="platform", slug="THIS-IS_a_SLUG_123")

    def test_none_if_not_found(self, tmp_path):
        file_contents = "window.location.replace(https://os.mbed.com/about);"
        pathlib.Path(tmp_path, "MBED.HTM").write_text(file_contents)

        assert read_device_files([tmp_path]).online_id is None


class TestExtractsJlinkData:
    def test_reads_board_slug(self, tmp_path):
        board_html = pathlib.Path(tmp_path, "Board.html")
        board_html.write_text(build_board_html("http://test.com/test/slug"))

        info = read_device_files([tmp_path])

        assert info.online_id == OnlineId(target_type="jlink", slug="slug")

    def test_reads_board_slug_ignore_extension(self, tmp_path):
        board_html = pathlib.Path(tmp_path, "Board.html")
        board_html.write_text(build_board_html("http://test.com/slug.html"))

        info = read_device_files([tmp_path])

        assert info.online_id == OnlineId(target_type="jlink", slug="slug")

    def test_id_none_if_no_board_slug(self, tmp_path):
        board_html = pathlib.Path(tmp_path, "Board.html")
        board_html.write_text(build_board_html("http://test.com"))

        info = read_device_files([tmp_path])

        assert info.online_id is None

    def test_reads_segger_slug(self, tmp_path):
        segger_html = pathlib.Path(tmp_path, "Segger.html")
        segger_html.write_text(build_segger_html("variant"))

        info = read_device_files([tmp_path])

        assert info.interface_details.get("Version") == "variant"

    def test_interface_empty_if_not_found(self, tmp_path):
        board_html = pathlib.Path(tmp_path, "Board.html")
        board_html.write_text("<html></html>")

        info = read_device_files([tmp_path])

        assert info.interface_details == {}


# Helpers to build test data
def build_short_details_txt(version="0226", build="Aug 24 2015 17:06:30", commit_sha="27a2367", local_mods="Yes"):
    return (
        f"""Version: {version}
Build:   {build}
Git Commit SHA: {commit_sha}
Git Local mods: {local_mods}
""",
        {"Version": version, "Build": build, "Git Commit SHA": commit_sha, "Git Local mods": local_mods},
    )


def build_long_details_txt(
    interface_version="0226",
    uid="0240000029164e45002f0012706e0006f301000097969900",
    hif_id="97969900",
    auto_reset="0",
    auto_allow="0",
    daplink_mode="Interface",
    commit_sha="c76899838",
    local_mods="0",
    usb_ifaces="MSD, CDC, HID",
    iface_crc="0x26764ebf",
):
    return (
        f"""# DAPLink Firmware - see https://mbed.com/daplink
Unique ID: {uid}
HIF ID: {hif_id}
Auto Reset: {auto_reset}
Automation allowed: {auto_allow}
Daplink Mode: {daplink_mode}
Interface Version: {interface_version}
Git SHA: {commit_sha}
Local Mods: {local_mods}
USB Interfaces: {usb_ifaces}
Interface CRC: {iface_crc}
""",
        {
            "Unique ID": uid,
            "HIF ID": hif_id,
            "Auto Reset": auto_reset,
            "Automation allowed": auto_allow,
            "Daplink Mode": daplink_mode,
            "Version": interface_version,
            "Git SHA": commit_sha,
            "Local Mods": local_mods,
            "USB Interfaces": usb_ifaces,
            "Interface CRC": iface_crc,
        },
    )


def build_segger_html(model="j-link-ob"):
    return (
        """<html><head><meta http-equiv="refresh" content="0;"""
        f"""url=https://www.segger.com/products/debug-probes/j-link/models/{model}/"/>"""
        """</head><body></body></html>"""
    )


def build_board_html(target_url="http://www.nxp.com/FRDM-K64F"):
    return f"""<html><head><meta http-equiv="refresh" content="0; url={target_url}"/></head><body></body></html>"""


class TestReadDetailsTxt:
    @pytest.mark.parametrize(
        "content, expected",
        (
            build_short_details_txt(),
            build_short_details_txt(version="0777", commit_sha="99789s", local_mods="No"),
            build_long_details_txt(),
            ("", {}),
            ("\n", {}),
            ("blablablablaandbla", {}),
            ("blablabla\nblaandbla\nversion : 2\n\n", {"version": "2"}),
        ),
        ids=("short", "short2", "long", "empty", "newline", "nosep", "multiline"),
    )
    def test_parses_details_txt(self, content, expected, tmp_path):
        details_file_path = pathlib.Path(tmp_path, "DETAILS.txt")
        details_file_path.write_text(content)

        result = read_device_files([tmp_path]).interface_details
        assert result == expected



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Parses files found on Mbed enabled devices.

There are a number of data files stored on an mbed enabled device's USB mass storage.

The schema and content of these files are described in detail below.

- MBED.HTM File
We support many flavours of MBED.HTM files. The list of examples below is not exhaustive.

    <!-- mbed Microcontroller Website and Authentication Shortcut -->
    <!-- Version: 0200 Build: Feb  3 2014 14:03:10 -->
    <html>
    <head>
    <meta http-equiv="refresh" content="0; url=http://mbed.org/device/?code=0710020092566DA9AD33FB84"/>
    <title>mbed Website Shortcut</title>
    </head>
    <body></body>
    </html>

---

    <!doctype html>
    <!-- mbed Platform Website and Authentication Shortcut -->
    <html>
    <head>
    <meta charset="utf-8">
    <title>mbed Website Shortcut</title>
    </head>
    <body>
    <script>
    window.location.replace("https://mbed.org/device/?code=460000000988254a00000000000000000000000097969902?version=0253?target_id=00000000000000000000000000000000");
    </script>
    </body>
    </html>

---

    <!-- mbed Microcontroller Website and Authentication Shortcut -->
    <html>
    <head>
    <meta http-equiv="refresh" content="0; url=http://mbed.org/start?auth=101000000000000000000002F7F35E602eeb0bb9b632205c51f6c357aeee7bc9&loader=11972&firmware=16457&configuration=4" />  # noqa
    <title>mbed Website Shortcut</title>
    </head>
    <body></body>
    </html>

---

    <!doctype html>
    <!-- mbed Platform Website and Authentication Shortcut -->
    <html>
    <head>
    <meta charset="utf-8">
    <title>mbed Website Shortcut</title>
    </head>
    <body>
    <script>
    window.location.replace("https://os.mbed.com/platforms/LPCXpresso54114/");
    </script>
    </body>
    </html>

---

- Segger.html File
This can be found on connected J-Link devices. An example is shown below.

    <html>
    <head>
    <meta http-equiv="refresh" content="0; url=https://www.segger.com/products/debug-probes/j-link/models/j-link-ob/"/>
    </head>
    <body></body>
    </html>
---

- Board.html File
This can be found on connected J-Link devices. An example is shown below.

    <html>
    <head>
    <meta http-equiv="refresh" content="0; url=http://www.nxp.com/FRDM-K64F"/>
    <title>NXP Product Page</title>
    </head>
    <body></body>
    </html>

"""
import logging
import pathlib
import re

from dataclasses import dataclass
from typing import Optional, NamedTuple, Iterable, List


logger = logging.getLogger(__name__)


class OnlineId(NamedTuple):
    """Used to identify the target against the os.mbed.com website.

    The target type and slug are used in the URI for the board and together they can be used uniquely identify a board.

    OnlineId(target_type="platform", slug="SOME-SLUG") -> https://os.mbed.com/platforms/SOME-SLUG
    """

    target_type: str
    slug: str


@dataclass
class DeviceFileInfo:
    """Information gathered from Mbed device files."""

    product_code: Optional[str]
    online_id: Optional[OnlineId]
    interface_details: dict


def read_device_files(directory_paths: Iterable[pathlib.Path]) -> DeviceFileInfo:
    """Read data from files contained on an mbed enabled device's USB mass storage device.

    If details.txt exists and it contains a product code, then we will use that code. If not then we try to grep the
    code from the mbed.htm file. We extract an OnlineID from mbed.htm as we also make use of that information to find a
    board entry in Mbed OS's various target databases and JSON files.

    Args:
        directory_paths: Paths to the directories containing device files.
    """
    device_file_paths = _get_device_file_paths(directory_paths)
    if not device_file_paths:
        paths = "\n".join(str(p) for p in directory_paths)
        logger.warning(
            f"No files were found in the device's mass storage device. The following paths were searched:\n{paths}."
            "\nThis device may not be identifiable as Mbed enabled. Check the files exist, are not hidden and are not "
            "corrupted."
        )
        return DeviceFileInfo(None, None, {})

    htm_file_contents = _read_htm_file_contents(device_file_paths)
    details_txt_contents = _read_first_details_txt_contents(device_file_paths)
    # details.txt is the "preferred" source of truth for the product_code
    code = details_txt_contents.get("code")
    if code is None:
        # erk! well, let's get it from the mbed.htm file instead...
        code = _extract_product_code_from_htm(htm_file_contents)

    online_id = _extract_online_id_from_htm(htm_file_contents)
    if online_id is None:
        # If no online ID available from .htm, may be a J-Link
        online_id = _extract_online_id_jlink_html(device_file_paths)
        details_txt_contents.update(_extract_version_jlink_html(device_file_paths))
    return DeviceFileInfo(code, online_id, details_txt_contents)


def _read_product_code(file_contents: str) -> Optional[str]:
    """Returns product code parsed from the file contents, None if not found."""
    regex = r"""
            (?:code|auth)=                   # attribute name
            (?P<product_code>[a-fA-F0-9]{4}) # product code
    """
    match = re.search(regex, file_contents, re.VERBOSE)
    if match:
        return match["product_code"]
    return None


def _read_online_id(file_contents: str) -> Optional[OnlineId]:
    """Returns online id parsed from the files contents, None if not found."""
    regex = r"""
            (?P<target_type>module|platform)s   # module|platform
            \/                                  # forward slash in the url
            (?P<slug>[-\w]+)                    # permitted characters in a slug are letters and digits
    """
    match = re.search(regex, file_contents, re.VERBOSE)
    if match:
        return OnlineId(target_type=match["target_type"], slug=match["slug"])
    return None


def _read_url_slug(file_contents: str) -> Optional[str]:
    """Returns slug parsed from file contents, None if not found."""
    regex = r"""url=[^"]+\.[^"]+\/(?P<slug>[-\w]+)(\/|[\w.]+)?\""""
    match = re.search(regex, file_contents, re.VERBOSE)
    if match:
        return match["slug"]
    return None


def _read_first_details_txt_contents(file_paths: Iterable[pathlib.Path]) -> dict:
    for path in file_paths:
        if _is_details_txt(path):
            contents = _try_read_file_text(path)
            if contents:
                return _read_details_txt(contents)

    logger.warning(f"Could not find DETAILS.TXT in {file_paths!r}.")
    return {}


def _read_details_txt(file_contents: str) -> dict:
    """Parse the contents of a daplink-compatible device's details.txt.

    Args:
        file_contents: The contents of the details.txt file.
    """
    output = {}
    for line in file_contents.splitlines():
        # Ignore any comments in the file contents
        if line.startswith("#"):
            continue

        key, sep, value = line.partition(":")
        if key and value:
            output[key.strip()] = value.strip()

    # Some forms of details.txt use Interface Version instead of Version as the key for the version number field
    if "Interface Version" in output and "Version" not in output:
        output["Version"] = output.pop("Interface Version")

    return output


def _extract_product_code_from_htm(all_files_contents: Iterable[str]) -> Optional[str]:
    """Return first product code found in files contents, None if not found."""
    for contents in all_files_contents:
        product_code = _read_product_code(contents)
        if product_code:
            return product_code
    return None


def _extract_online_id_from_htm(all_files_contents: Iterable[str]) -> Optional[OnlineId]:
    """Return first online ID found in files contents, None if not found."""
    for contents in all_files_contents:
        online_id = _read_online_id(contents)
        if online_id:
            return online_id
    return None


def _extract_online_id_jlink_html(file_paths: Iterable[pathlib.Path]) -> Optional[OnlineId]:
    """Return online ID found in Board.html, None if not found."""
    contents = _get_board_html_contents(file_paths)
    if contents is not None:
        slug = _read_url_slug(contents)
        if slug:
            return OnlineId("jlink", slug)
    return None


def _extract_version_jlink_html(file_paths: Iterable[pathlib.Path]) -> dict:
    """Return dict with version found in Segger.html, empty if not found."""
    interface_data = {}
    contents = _get_segger_html_content(file_paths)
    if contents is not None:
        segger_version = _read_url_slug(contents)
        if segger_version:
            interface_data["Version"] = segger_version
    return interface_data


def _get_device_file_paths(directories: Iterable[pathlib.Path]) -> List[pathlib.Path]:
    return [path for directory in directories for path in directory.iterdir() if not _is_hidden_file(path)]


def _try_read_file_text(file_path: pathlib.Path) -> Optional[str]:
    try:
        return file_path.read_text()
    except OSError:
        logger.warning(f"The file '{file_path}' could not be read from the device, target may not be identified.")
        return None


def _read_htm_file_contents(all_files: Iterable[pathlib.Path]) -> List[str]:
    htm_files_contents = []
    for file in all_files:
        if _is_htm_file(file):
            contents = _try_read_file_text(file)
            if contents:
                htm_files_contents.append(contents)
    return htm_files_contents


def _get_segger_html_content(file_paths: Iterable[pathlib.Path]) -> Optional[str]:
    for fp in file_paths:
        if fp.name.lower() == "segger.html":
            return _try_read_file_text(fp)
    return None


def _get_board_html_contents(file_paths: Iterable[pathlib.Path]) -> Optional[str]:
    for fp in file_paths:
        if fp.name.lower() == "board.html":
            return _try_read_file_text(fp)
    return None


def _is_hidden_file(file: pathlib.Path) -> bool:
    """Checks if the file is hidden."""
    return file.name.startswith(".")


def _is_htm_file(path: pathlib.Path) -> bool:
    return path.suffix.lower() == ".htm"


def _is_details_txt(path: pathlib.Path) -> bool:
    return path.name.lower() == "details.txt"



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from pathlib import Path
from unittest import mock

from mbed_tools.project._internal.render_templates import (
    render_cmakelists_template,
    render_main_cpp_template,
    render_gitignore_template,
)


@mock.patch("mbed_tools.project._internal.render_templates.datetime")
class TestRenderTemplates:
    def test_renders_cmakelists_template(self, mock_datetime, tmp_path):
        the_year = 3999
        mock_datetime.datetime.now.return_value.year = the_year
        program_name = "mytestprogram"
        file_path = Path(tmp_path, "mytestpath")

        render_cmakelists_template(file_path, program_name)
        output = file_path.read_text()

        assert str(the_year) in output
        assert program_name in output

    def test_renders_main_cpp_template(self, mock_datetime, tmp_path):
        the_year = 3999
        mock_datetime.datetime.now.return_value.year = the_year
        file_path = Path(tmp_path, "mytestpath")

        render_main_cpp_template(file_path)

        assert str(the_year) in file_path.read_text()

    def test_renders_gitignore_template(self, _, tmp_path):
        file_path = Path(tmp_path, "mytestpath")

        render_gitignore_template(file_path)

        assert "cmake_build" in file_path.read_text()
        assert ".mbedbuild" in file_path.read_text()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for `mbed_tools.targets.get_board`."""
import pytest

from unittest import mock

from mbed_tools.targets._internal.exceptions import BoardAPIError

# Import from top level as this is the expected interface for users
from mbed_tools.targets import get_board_by_online_id, get_board_by_product_code, get_board_by_jlink_slug
from mbed_tools.targets.get_board import (
    _DatabaseMode,
    _get_database_mode,
    get_board,
)
from mbed_tools.targets.env import env
from mbed_tools.targets.exceptions import UnknownBoard, UnsupportedMode
from tests.targets.factories import make_board


@pytest.fixture
def mock_get_board():
    with mock.patch("mbed_tools.targets.get_board.get_board", autospec=True) as gbp:
        yield gbp


@pytest.fixture
def mock_env():
    with mock.patch("mbed_tools.targets.get_board.env", spec_set=env) as gbp:
        yield gbp


@pytest.fixture
def mocked_boards():
    with mock.patch("mbed_tools.targets.get_board.Boards", autospec=True) as gbp:
        yield gbp


class TestGetBoard:
    def test_online_mode(self, mock_env, mocked_boards):
        mock_env.MBED_DATABASE_MODE = "ONLINE"
        fn = mock.Mock()

        subject = get_board(fn)

        assert subject == mocked_boards.from_online_database().get_board.return_value
        mocked_boards.from_online_database().get_board.assert_called_once_with(fn)

    def test_offline_mode(self, mock_env, mocked_boards):
        mock_env.MBED_DATABASE_MODE = "OFFLINE"
        fn = mock.Mock()

        subject = get_board(fn)

        assert subject == mocked_boards.from_offline_database().get_board.return_value
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)

    def test_auto_mode_calls_offline_boards_first(self, mock_env, mocked_boards):
        mock_env.MBED_DATABASE_MODE = "AUTO"
        fn = mock.Mock()

        subject = get_board(fn)

        assert subject == mocked_boards.from_offline_database().get_board.return_value
        mocked_boards.from_online_database().get_board.assert_not_called()
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)

    def test_auto_mode_falls_back_to_online_database_when_board_not_found(self, mock_env, mocked_boards):
        mock_env.MBED_DATABASE_MODE = "AUTO"
        mocked_boards.from_offline_database().get_board.side_effect = UnknownBoard
        fn = mock.Mock()

        subject = get_board(fn)

        assert subject == mocked_boards.from_online_database().get_board.return_value
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)
        mocked_boards.from_online_database().get_board.assert_called_once_with(fn)

    def test_auto_mode_raises_when_board_not_found_offline_with_no_network(self, mock_env, mocked_boards):
        mock_env.MBED_DATABASE_MODE = "AUTO"
        mocked_boards.from_offline_database().get_board.side_effect = UnknownBoard
        mocked_boards.from_online_database().get_board.side_effect = BoardAPIError
        fn = mock.Mock()

        with pytest.raises(UnknownBoard):
            get_board(fn)
        mocked_boards.from_offline_database().get_board.assert_called_once_with(fn)
        mocked_boards.from_online_database().get_board.assert_called_once_with(fn)


class TestGetBoardByProductCode:
    def test_matches_boards_by_product_code(self, mock_get_board):
        product_code = "swag"

        assert get_board_by_product_code(product_code) == mock_get_board.return_value

        # Test callable matches correct boards
        fn = mock_get_board.call_args[0][0]

        matching_board = make_board(product_code=product_code)
        not_matching_board = make_board(product_code="whatever")

        assert fn(matching_board)
        assert not fn(not_matching_board)


class TestGetBoardByOnlineId:
    def test_matches_boards_by_online_id(self, mock_get_board):
        target_type = "platform"

        assert get_board_by_online_id(slug="slug", target_type=target_type) == mock_get_board.return_value

        # Test callable matches correct boards
        fn = mock_get_board.call_args[0][0]

        matching_board_1 = make_board(target_type=target_type, slug="slug")
        matching_board_2 = make_board(target_type=target_type, slug="SlUg")
        not_matching_board = make_board(target_type=target_type, slug="whatever")

        assert fn(matching_board_1)
        assert fn(matching_board_2)
        assert not fn(not_matching_board)


class TestGetBoardByJlinkSlug:
    def test_matches_boards_by_online_id(self, mock_get_board):
        assert get_board_by_jlink_slug(slug="slug") == mock_get_board.return_value

        # Test callable matches correct boards
        fn = mock_get_board.call_args[0][0]

        matching_board_1 = make_board(slug="slug")
        matching_board_2 = make_board(board_type="slug")
        matching_board_3 = make_board(board_name="slug")
        not_matching_board = make_board()

        assert fn(matching_board_1)
        assert fn(matching_board_2)
        assert fn(matching_board_3)
        assert not fn(not_matching_board)


class TestGetDatabaseMode:
    def test_returns_configured_database_mode(self, mock_env):
        mock_env.MBED_DATABASE_MODE = "OFFLINE"
        assert _get_database_mode() == _DatabaseMode.OFFLINE

    def test_raises_when_configuration_is_not_supported(self, mock_env):
        mock_env.MBED_DATABASE_MODE = "NOT_VALID"
        with pytest.raises(UnsupportedMode):
            _get_database_mode()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Data model definition for Device and ConnectedDevices."""
from dataclasses import dataclass, field
from pathlib import Path
from typing import Tuple, Optional, List
from mbed_tools.targets import Board
from mbed_tools.devices._internal.detect_candidate_devices import CandidateDevice
from mbed_tools.devices._internal.resolve_board import resolve_board, NoBoardForCandidate, ResolveBoardError
from mbed_tools.devices._internal.file_parser import read_device_files
from mbed_tools.devices.exceptions import DeviceLookupFailed


@dataclass(frozen=True, order=True)
class Device:
    """Definition of an Mbed Enabled Device.

    An Mbed Device is always a USB mass storage device, which sometimes also presents a USB serial port.
    A valid Mbed Device must have a Board associated with it.

    Attributes:
        mbed_board: The Board associated with this device.
        serial_number: The serial number presented by the device to the USB subsystem.
        serial_port: The serial port presented by this device, could be None.
        mount_points: The filesystem mount points associated with this device.
    """

    mbed_board: Board
    serial_number: str
    serial_port: Optional[str]
    mount_points: Tuple[Path, ...]
    mbed_enabled: bool = False
    interface_version: Optional[str] = None

    @classmethod
    def from_candidate(cls, candidate: CandidateDevice) -> "Device":
        """Contruct a Device from a CandidateDevice.

        We try to resolve a board using data files that may be stored on the CandidateDevice.
        If this fails we set the board to `None` which means we couldn't verify this Device
        as being an Mbed enabled device.

        Args:
            candidate: The CandidateDevice we're using to create the Device.
        """
        device_file_info = read_device_files(candidate.mount_points)
        try:
            mbed_board = resolve_board(
                device_file_info.product_code, device_file_info.online_id, candidate.serial_number
            )
            mbed_enabled = True
        except NoBoardForCandidate:
            # Create an empty Board to ensure the device is fully populated and rendering is simple
            mbed_board = Board.from_offline_board_entry({})
            mbed_enabled = False
        except ResolveBoardError:
            raise DeviceLookupFailed(
                f"Failed to resolve the board for candidate device {candidate!r}. There was a problem looking up the "
                "board data in the database."
            )

        return Device(
            serial_port=candidate.serial_port,
            serial_number=candidate.serial_number,
            mount_points=candidate.mount_points,
            mbed_board=mbed_board,
            mbed_enabled=mbed_enabled,
            interface_version=device_file_info.interface_details.get("Version"),
        )


@dataclass(order=True)
class ConnectedDevices:
    """Definition of connected devices which may be Mbed Boards.

    If a connected device is identified as an Mbed Board by using the HTM file on the USB mass storage device (or
    sometimes by using the serial number), it will be included in the `identified_devices` list.

    However, if the device appears as if it could be an Mbed Board but it has not been possible to find a matching
    entry in the database then it will be included in the `unidentified_devices` list.

    Attributes:
        identified_devices: A list of devices that have been identified as MbedTargets.
        unidentified_devices: A list of devices that could potentially be MbedTargets.
    """

    identified_devices: List[Device] = field(default_factory=list)
    unidentified_devices: List[Device] = field(default_factory=list)

    def add_device(self, device: Device) -> None:
        """Add a device to the connected devices.

        Args:
            device: a Device object containing the device information.
        """
        if not device.mbed_enabled:
            # Keep a list of devices that could not be identified but are Mbed Boards
            self.unidentified_devices.append(device)
        else:
            # Keep a list of devices that have been identified as Mbed Boards
            self.identified_devices.append(device)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Build configuration representation."""
import logging

from collections import UserDict
from typing import Any, Iterable, Hashable, List

from mbed_tools.build._internal.config.source import Override, ConfigSetting

logger = logging.getLogger(__name__)


class Config(UserDict):
    """Mapping of config settings.

    This object understands how to populate the different 'config sections' which all have different rules for how the
    settings are collected.
    Applies overrides, appends macros, and updates config settings.
    """

    def __setitem__(self, key: Hashable, item: Any) -> None:
        """Set an item based on its key."""
        if key == CONFIG_SECTION:
            self._update_config_section(item)
        elif key == OVERRIDES_SECTION:
            self._handle_overrides(item)
        elif key == MACROS_SECTION:
            self.data[MACROS_SECTION] = self.data.get(MACROS_SECTION, set()) | item
        elif key == REQUIRES_SECTION:
            self.data[REQUIRES_SECTION] = self.data.get(REQUIRES_SECTION, set()) | item
        else:
            super().__setitem__(key, item)

    def _handle_overrides(self, overrides: Iterable[Override]) -> None:
        for override in overrides:
            logger.debug("Applying override '%s.%s'='%s'", override.namespace, override.name, repr(override.value))
            if override.name in self.data:
                _apply_override(self.data, override)
                continue

            setting = next(
                filter(
                    lambda x: x.name == override.name and x.namespace == override.namespace,
                    self.data.get(CONFIG_SECTION, []),
                ),
                None,
            )
            if setting is None:
                logger.warning(
                    f"You are attempting to override an undefined config parameter "
                    f"`{override.namespace}.{override.name}`.\n"
                    "It is an error to override an undefined configuration parameter. "
                    "Please check your target_overrides are correct.\n"
                    f"The parameter `{override.namespace}.{override.name}` will not be added to the configuration."
                )
            else:
                setting.value = override.value

    def _update_config_section(self, config_settings: List[ConfigSetting]) -> None:
        for setting in config_settings:
            logger.debug("Adding config setting: '%s.%s'", setting.namespace, setting.name)
            if setting in self.data.get(CONFIG_SECTION, []):
                raise ValueError(
                    f"Setting {setting.namespace}.{setting.name} already defined. You cannot duplicate config settings!"
                )

        self.data[CONFIG_SECTION] = self.data.get(CONFIG_SECTION, []) + config_settings


CONFIG_SECTION = "config"
MACROS_SECTION = "macros"
OVERRIDES_SECTION = "overrides"
REQUIRES_SECTION = "requires"


def _apply_override(data: dict, override: Override) -> None:
    if override.modifier == "add":
        data[override.name] |= override.value
    elif override.modifier == "remove":
        data[override.name] -= override.value
    else:
        data[override.name] = override.value



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import mock

import pytest

from mbed_tools.targets.exceptions import UnknownBoard, MbedTargetsError

from mbed_tools.devices._internal.file_parser import OnlineId, DeviceFileInfo
from mbed_tools.devices._internal.resolve_board import NoBoardForCandidate, resolve_board, ResolveBoardError


@pytest.fixture
def get_board_by_product_code_mock():
    with mock.patch("mbed_tools.devices._internal.resolve_board.get_board_by_product_code", autospec=True) as gbp:
        yield gbp


@pytest.fixture
def get_board_by_jlink_slug_mock():
    with mock.patch("mbed_tools.devices._internal.resolve_board.get_board_by_jlink_slug", autospec=True) as gbp:
        yield gbp


@pytest.fixture
def get_board_by_online_id_mock():
    with mock.patch("mbed_tools.devices._internal.resolve_board.get_board_by_online_id", autospec=True) as gbp:
        yield gbp


class TestResolveBoardUsingProductCodeFromHTM:
    def test_returns_resolved_target(self, get_board_by_product_code_mock):
        dev_info = DeviceFileInfo("0123", None, None)

        subject = resolve_board(product_code=dev_info.product_code)

        assert subject == get_board_by_product_code_mock.return_value
        get_board_by_product_code_mock.assert_called_once_with(dev_info.product_code)

    def test_raises_when_board_not_found(self, get_board_by_product_code_mock):
        get_board_by_product_code_mock.side_effect = UnknownBoard

        with pytest.raises(NoBoardForCandidate):
            resolve_board(product_code="0123")

    def test_raises_when_database_lookup_fails(self, get_board_by_product_code_mock, caplog):
        get_board_by_product_code_mock.side_effect = MbedTargetsError
        product_code = "0123"

        with pytest.raises(ResolveBoardError):
            resolve_board(product_code=product_code)

        assert product_code in caplog.text
        assert caplog.records[-1].levelname == "ERROR"


class TestResolveBoardUsingOnlineIdFromHTM:
    def test_returns_resolved_board(self, get_board_by_online_id_mock):
        online_id = OnlineId(target_type="hat", slug="boat")

        subject = resolve_board(online_id=online_id)

        assert subject == get_board_by_online_id_mock.return_value
        get_board_by_online_id_mock.assert_called_once_with(target_type=online_id.target_type, slug=online_id.slug)

    def test_raises_when_board_not_found(self, get_board_by_online_id_mock):
        get_board_by_online_id_mock.side_effect = UnknownBoard

        with pytest.raises(NoBoardForCandidate):
            resolve_board(online_id=OnlineId(target_type="hat", slug="boat"))

    def test_raises_when_database_lookup_fails(self, get_board_by_online_id_mock, caplog):
        get_board_by_online_id_mock.side_effect = MbedTargetsError
        online_id = OnlineId(target_type="hat", slug="boat")

        with pytest.raises(ResolveBoardError):
            resolve_board(online_id=online_id)

        assert repr(online_id) in caplog.text
        assert caplog.records[-1].levelname == "ERROR"


class TestResolveBoardUsingSlugFromJlink:
    def test_returns_resolved_board(self, get_board_by_jlink_slug_mock):
        online_id = OnlineId("jlink", "test-board")

        subject = resolve_board(online_id=online_id)

        assert subject == get_board_by_jlink_slug_mock.return_value
        get_board_by_jlink_slug_mock.assert_called_once_with(online_id.slug)

    def test_raises_when_board_not_found(self, get_board_by_jlink_slug_mock):
        get_board_by_jlink_slug_mock.side_effect = UnknownBoard
        online_id = OnlineId("jlink", "test-board")

        with pytest.raises(NoBoardForCandidate):
            resolve_board(online_id=online_id)

    def test_raises_when_database_lookup_fails(self, get_board_by_jlink_slug_mock, caplog):
        get_board_by_jlink_slug_mock.side_effect = MbedTargetsError
        online_id = OnlineId("jlink", "test-board")

        with pytest.raises(ResolveBoardError):
            resolve_board(online_id=online_id)

        assert repr(online_id) in caplog.text
        assert caplog.records[-1].levelname == "ERROR"


class TestResolveBoardUsingProductCodeFromSerial:
    def test_resolves_board_using_product_code_when_available(self, get_board_by_product_code_mock):
        serial_number = "0A9KJFKD0993WJKUFS0KLJ329090"
        subject = resolve_board(serial_number=serial_number)

        assert subject == get_board_by_product_code_mock.return_value
        get_board_by_product_code_mock.assert_called_once_with(serial_number[:4])

    def test_raises_when_board_not_found(self, get_board_by_product_code_mock):
        get_board_by_product_code_mock.side_effect = UnknownBoard

        with pytest.raises(NoBoardForCandidate):
            resolve_board(serial_number="0")

    def test_raises_when_database_lookup_fails(self, get_board_by_product_code_mock, caplog):
        get_board_by_product_code_mock.side_effect = MbedTargetsError
        serial_number = "0123456"

        with pytest.raises(ResolveBoardError):
            resolve_board(serial_number=serial_number)

        assert serial_number[:4] in caplog.text
        assert caplog.records[-1].levelname == "ERROR"
